#!/usr/bin/env python

from sss import ssslib
import sys
import re
import string
import xml.dom.minidom

def cmd_mappings(command,component):
    if component == "build-system":
        return {"get":"get-node-software",
                "set":"set-node-software",
                "add":"add-node-software",
                "del":"del-node-software"}[command]
    if component == "cluster-hardware":
        return {"get":"get-node-hardware",
                "set":"set-node-hardware",
                "add":"add-node-hardware",
                "del":"del-node-hardware"}[command]
    if component == "node-state-manager":
        return {"get":"get-node-state",
                "set":"set-node-state",
                "add":"add-node-state",
                "del":"del-node-state"}[command]
    raise "NoCommand", (command,component)

def record_mappings(command,component):
    if component == "build-system":
        return "node-software"
    if component == "cluster-hardware":
        return "node-hardware"
    if component == "node-state-manager":
        return "node-state"
    raise "NoRecord", (command,component)

if __name__ == "__main__":

    cmdlist = ['get','set','add','del']
    cmdargs = {"get":"r","set":"sr","add":"s","del":"r"}
    argmtch = [["x","s"],["r","sr"]]

    def usage():
        print "Usage:"
        print "omniclient.py [-(d|n|p)] <command> [-s value pairs] [-r restriction pairs]"
        print "Where <command> is one of %s."%string.join(cmdlist,",")
        print "\t\t-d\tSSSLib debugging on."
        print "\t\t-n\tDryrun - describe actions that would be taken."
        print "\t\t-p\tPrint - Print nodes affected/retrieved, one per line."
        sys.exit(1)
    
    argre = re.compile("^(-[dnp] )*(?P<command>\w+) (-s (?P<values>(\w+=\w+ {0,1})+)){0,1} {0,1}(-r (?P<restrictions>(\w+=\w+ {0,1})+)){0,1}$")
    argstring = string.join(sys.argv[1:]," ")

    match = argre.match(argstring)
    if not match:
        usage()

    command = match.group("command")
    values = match.group("values")
    restrictions = match.group("restrictions")

    if not command in cmdlist:
        usage()

    if not cmdargs[command] == argmtch[not not restrictions][not not values]:
        usage()

    bs_handled = ['build-system','class','mayor','image','kernel','kernel_opts','action']
    ch_handled = ['cluster-hardware','powerdev','powerport','consoledev','consoleport','switch','switchport','hwaddr','iftype']
    ns_handled = ['node-state-manager','state','adminstate']
    
    req_info = {"build-system":[],
                "cluster-hardware":[],
                "node-state-manager":[]}

    val_info = {"build-system":[],
                "cluster-hardware":[],
                "node-state-manager":[]}

    req_keyvals = {}
    val_keyvals = {}

    # make the key->val mappings that represent the action -s and -r values
    # also make the component->[keys to be requested from that component] mappings
    
    if restrictions:
        for keyval in string.split(restrictions):
            key,val = string.split(keyval,"=")
            if not key == "name":
                component = filter(lambda x:key in x,[bs_handled,ch_handled,ns_handled])[0][0]
                req_info[component].append(key)
            else:
                for element in req_info.keys():
                    req_info[element].append("name")
            req_keyvals[key] = val

    if values:
        if command == "add" and not "name" in map(lambda x:string.split(x,"=")[0],string.split(values)):
            print "Warning: adding a record for a node with no name, will be entered as unnamed_node in the databases."
        for keyval in string.split(values):
            key,val = string.split(keyval,"=")
            if not key == "name":
                component = filter(lambda x:key in x,[bs_handled,ch_handled,ns_handled])[0][0]
                val_info[component].append((key,val))
            else:
                for element in val_info.keys():
                    val_info[element].append(("name",val))
            val_keyvals[key] = val

    nodelist = {}

    sss = ssslib.comm_lib(debug=("-d" in sys.argv[1:]))

    # generate the initial nodelist from those components with restrictions applied to them

    # the nodelist is a mapping from node name to list of information elements from the various components

    for (component,keys) in filter(lambda x:len(x[1]),req_info.iteritems()):
        req_doc = xml.dom.minidom.Document()
        req_root = req_doc.createElement(cmd_mappings("get",component))
        req_record = req_doc.createElement(record_mappings("get",component))
        for key in keys:
            req_record.setAttribute(key,req_keyvals[key])
        req_root.appendChild(req_record)
        handle = sss.ClientInit(component)
        sss.SendMessage(handle,req_root.toxml())
        msg = sss.RecvMessage(handle)
        sss.ClientClose(handle)
        msg_obj = xml.dom.minidom.parseString(msg)
        for child in msg_obj.documentElement.childNodes:
            if child.getAttribute("name") in nodelist.keys():
                nodelist[child.getAttribute("name")].append(child)
            else:
                nodelist[child.getAttribute("name")] = [child]
    
    nodelist2 = {}

    # filter out everything that wasn't in all requested databases for which restrictions were applied (except name - a special case)

    #map(lambda x:nodelist2.__setitem__(x[0],x[1]),filter(lambda x:len(x[1]) == len(filter(lambda x:len(x[1]),req_info.iteritems())),nodelist.iteritems()))

    map(lambda x:nodelist2.__setitem__(x[0],x[1]),filter(lambda x:len(x[1]) == len(filter(lambda x:len(x[1]),req_info.iteritems())),nodelist.iteritems()))

    nodelist = nodelist2

    # ask all other components to add their data on the nodes in the initial list

    for component,None in filter(lambda x:not len(x[1]),req_info.iteritems()):
        req_doc = xml.dom.minidom.Document()
        req_root = req_doc.createElement(cmd_mappings("get",component))
        req_record = req_doc.createElement(record_mappings("get",component))
        req_root.appendChild(req_record)
        handle = sss.ClientInit(component)
        sss.SendMessage(handle,req_root.toxml())
        msg = sss.RecvMessage(handle)
        sss.ClientClose(handle)
        msg_obj = xml.dom.minidom.parseString(msg)
        for child in msg_obj.documentElement.childNodes:
            if child.getAttribute("name") in nodelist.keys():
                nodelist[child.getAttribute("name")].append(child)
    
    nodeinfo = {}

    # construct a node->(xml with filled-out key->val attributes) mapping

    for (name,xml_list) in nodelist.iteritems():
        doc = xml.dom.minidom.Document()
        nodeinfo[name] = doc.createElement(name)
        for xml_el in xml_list:
            for (key,val) in xml_el._get_attributes().items():
                nodeinfo[name].setAttribute(key,val)

    #for (component,(key,val)) in val_info.iteritems():
    #val_doc = xml.dom.minidom.Document()
    #val_root = val_doc.createElement(cmd_mappings(command,component))

    # now we have a list of all the nodes we are dealing with
    # do more specific command-handling

    if command == "get":
        # we're done already
        if '-p' in sys.argv:
            for name,el in nodeinfo.iteritems():
                print name
        else:
            for name,el in nodeinfo.iteritems():
                print "--------- %s ---------"%name
                for a in el._get_attributes().items():
                    print "%s: %s"%a
        
    elif command == "set":
        # we have the nodelist, and the component->handled key buckets(val_info)

        # go through val_info, passing nodelist regexp and keyval pairs
        
        nodelist_re = string.join(nodeinfo.keys(),"|")
        namelist = []

        for (component,keys) in filter(lambda x:len(x[1]),val_info.iteritems()):
            req_doc = xml.dom.minidom.Document()
            req_root = req_doc.createElement(cmd_mappings(command,component))
            req_record = req_doc.createElement(record_mappings(command,component))
            for key in keys:
                req_root.setAttribute(key[0],key[1])
            req_record.setAttribute("name",nodelist_re)
            req_root.appendChild(req_record)
            handle = sss.ClientInit(component)
            sss.SendMessage(handle,req_root.toxml())
            msg = sss.RecvMessage(handle)
            sss.ClientClose(handle)
            msg_obj = xml.dom.minidom.parseString(msg)
            for child in msg_obj.documentElement.childNodes:
                if not child.getAttribute("name") in namelist:
                    namelist.append(child.getAttribute("name"))

        if "-p" in sys.argv:
            print string.join(namelist,"\n")

    elif command == "add":
        # for this command, we only care about the values, no restrictions

        namelist = []

        for (component,keys) in val_info.iteritems():
            req_doc = xml.dom.minidom.Document()
            req_root = req_doc.createElement(cmd_mappings(command,component))
            req_record = req_doc.createElement(record_mappings(command,component))
            for key in keys:
                req_record.setAttribute(key[0],key[1])
            req_root.appendChild(req_record)
            handle = sss.ClientInit(component)
            sss.SendMessage(handle,req_root.toxml())
            msg = sss.RecvMessage(handle)
            sss.ClientClose(handle)
            msg_obj = xml.dom.minidom.parseString(msg)
            for child in msg_obj.documentElement.childNodes:
                if not child.getAttribute("name") in namelist:
                    namelist.append(child.getAttribute("name"))

        if "-p" in sys.argv:
            print string.join(namelist,"\n")

    elif command == "del":
        # only care about restrictions, not values

        namelist = []

        for (component,keys) in req_info.iteritems():
            req_doc = xml.dom.minidom.Document()
            req_root = req_doc.createElement(cmd_mappings(command,component))
            req_record = req_doc.createElement(record_mappings(command,component))
            for key in keys:
                req_record.setAttribute(key,req_keyvals[key])
            req_root.appendChild(req_record)
            handle = sss.ClientInit(component)
            sss.SendMessage(handle,req_root.toxml())
            msg = sss.RecvMessage(handle)
            sss.ClientClose(handle)
            msg_obj = xml.dom.minidom.parseString(msg)
            for child in msg_obj.documentElement.childNodes:
                if not child.getAttribute("name") in namelist:
                    namelist.append(child.getAttribute("name"))

        if "-p" in sys.argv:
            print string.join(namelist,"\n")

    
