#!/usr/bin/env python

from os import popen3, system, unlink
from select import error as selecterror
from sss.node import Node, Msgbuf
from sss.server import SSSSelectServer, AddEvent
from sss.ssslib import ConnectError
from string import join,split
from syslog import syslog, LOG_INFO, LOG_ERR
from tempfile import mktemp
from threading import Thread
from xml.dom.minidom import Document
from time import time
from select import select

mpdpath='/usr/mpi/bin'

class Arg(Node):
    __attrfields__=['idx','value']

class Env(Node):
    __attrfields__=['name','value']    

class HostSpec(Msgbuf):
    pass

class ProcessSpec(Node):
    __attrfields__=['cwd','exec','path','user']
    typemap={'arg':Arg,'env':Env,'host-spec':HostSpec}

class ProcessGroup(Node):
    __attrfields__=['submitter','output','totalprocs','state']
    typemap={'process-spec':ProcessSpec}

    def __setdefaults__(self):
        setattr(self,'host-spec','')

    def MPIRun(self):
        # Dump out xml file
        tempfile=mktemp()
        self.tagname="create-process-group"
        open(tempfile,'w').write(self.toxmldom().toxml())
        self.tagname="process-group"
        syslog(LOG_INFO, "running :%s:"%("%s/mpdrun -f %s"%(mpdpath,tempfile)))
        (foo,out,err)=popen3("%s/mpdrun -f %s"%(mpdpath,tempfile))
        #unlink(tempfile)
        self.output = ""
        self.error  = ""
        out_should_exit = 0
        err_should_exit = 0
        while 1:
            socklist = []
            if not out_should_exit:
                socklist.append(out)
            if not err_should_exit:
                socklist.append(err)
            [reads,writes,excepts] = select(socklist,[],[])
            for r in reads:
                str = r.read(1)
                if r.fileno() == out.fileno():
                    self.output += str
                    if not len(str):
                        out_should_exit = 1
                if r.fileno() == err.fileno():
                    self.error += str
                    if not len(str):
                        err_should_exit = 1
            if out_should_exit and err_should_exit:
                syslog(LOG_INFO,"output was :%s:"%self.output)
                syslog(LOG_INFO,"error was :%s:"%self.error)
                return
 
    def Run(self):
        self.tagname="process-group"
        self.__attrfields__.append('pgid')
        self.state="STARTED"
        syslog(LOG_INFO,"PGid %s Started"%(self.pgid))
        try:
            AddEvent('process-manager','process_start',self.pgid)
        except ConnectError:
            syslog(LOG_ERR, "Error sending process_start for pgid %s"%(self.pgid))
        # MPIRun
        self.MPIRun()
        syslog(LOG_INFO,"PGid %s Exited"%(self.pgid))
        try:
            AddEvent('process-manager','process_end',self.pgid)
        except (ConnectError,ResponseError):
            syslog(LOG_ERR, "Error sending process_end for pgid %s"%(self.pgid))
        # Send out completion
        self.state="FINISHED"
        return None
    
    def Signal(self,signal):
        system("%s/mpdsigjob %s -a %s"%(mpdpath,signal,self.pgid))
        syslog(LOG_INFO,"PGid %s signaled with %s"%(self.pgid,signal))

    def toxmldom(self,spec=None):
        initdom=Node.toxmldom(self,spec)
        doc=Document()
        if hasattr(self,"output"):
            if self.output:
                out=doc.createElement("output")
                data=doc.createTextNode(self.output)
                out.appendChild(data)
                initdom.appendChild(out)
        if hasattr(self,"error"):
            if self.error:
                err=doc.createElement("error")
                data=doc.createTextNode(self.error)
                err.appendChild(data)
                initdom.appendChild(err)
        return initdom

class P4_ProcessGroup(Node):
    __attrfields__=['submitter','output','totalprocs']
    __textfields__=['host-spec']
    typemap={'process-spec':ProcessSpec}

    def MPIRun(self):
        # Dump out xml file
        tempfile=mktemp()
        syslog(LOG_INFO,"my host-spec is :%s:"%getattr(self,"host-spec"))
        open(tempfile,'w').write(getattr(self,"host-spec"))
        self.tagname="process-group"
        cmdline = 'su - %s -c "%s/mpirun -nolocal -np %s -machinefile %s'%(self.submitter,mpdpath,self.totalprocs,tempfile)
        cmdline += ' %s '%getattr(self.children[0],"exec")
        for arg in filter(lambda x:isinstance(x,Arg),self.children[0].children):
            cmdline += ' %s '%arg.value
        cmdline += '"'
        syslog(LOG_INFO, "running :%s:"%cmdline)
        (foo,out,err)=popen3(cmdline)
        #unlink(tempfile)
        self.output = ""
        self.error  = ""
        out_should_exit = 0
        err_should_exit = 0
        while 1:
            socklist = []
            if not out_should_exit:
                socklist.append(out)
            if not err_should_exit:
                socklist.append(err)
            [reads,writes,excepts] = select(socklist,[],[])
            for r in reads:
                str = r.read(1)
                if r.fileno() == out.fileno():
                    self.output += str
                    if not len(str):
                        out_should_exit = 1
                if r.fileno() == err.fileno():
                    self.error += str
                    if not len(str):
                        err_should_exit = 1
            if out_should_exit and err_should_exit:
                syslog(LOG_INFO,"output was :%s:"%self.output)
                syslog(LOG_INFO,"error was :%s:"%self.error)
                return
 
    def Run(self):
        self.tagname="process-group"
        self.__attrfields__.append('pgid')
        self.state="STARTED"
        syslog(LOG_INFO,"PGid %s Started"%(self.pgid))
        try:
            AddEvent('process-manager','process_start',self.pgid)
        except ConnectError:
            syslog(LOG_ERR, "Error sending process_start for pgid %s"%(self.pgid))
        # MPIRun
        self.MPIRun()
        syslog(LOG_INFO,"PGid %s Exited"%(self.pgid))
        try:
            AddEvent('process-manager','process_end',self.pgid)
        except ConnectError:
            syslog(LOG_ERR, "Error sending process_end for pgid %s"%(self.pgid))
        # Send out completion
        self.state="FINISHED"
        return None
    
    def Signal(self,signal):
        pass
        #system("%s/mpdsigjob %s -a %s"%(mpdpath,signal,self.pgid))
        #syslog(LOG_INFO,"PGid %s signaled with %s"%(self.pgid,signal))

    def toxmldom(self,spec=None):
        initdom=Node.toxmldom(self,spec)
        doc=Document()
        if hasattr(self,"output"):
            out=doc.createElement("output")
            data=doc.createTextNode(self.output)
            out.appendChild(data)
            initdom.appendChild(out)
        if hasattr(self,"error"):
            err=doc.createElement("error")
            data=doc.createTextNode(self.error)
            err.appendChild(data)
            initdom.appendChild(err)
        return initdom

class mpd(SSSSelectServer):
    __implementation__='mpdpm'
    __component__='process-manager'
    
    def __parseinput__(self):
        self.dispatch={"create-process-group":self.HandleXMLCreatePG,
                       "signal-process-group":self.HandleXMLPGSignal,
                       "kill-process-group":self.HandleXMLPGKill,
                       "get-process-group-info":self.HandleXMLPGInfo}
        self.pg={}
        self.threads=[]

    def GetPGID(self):
        return str(time())

    def ServeForever(self):
        while not self.shut:
            try:
                self.Serve()
                # reap dead threads
                map(lambda x:x.join(0), self.threads)
                live=filter(lambda x:x.isAlive(),self.threads)
                for t in [x for x in self.threads if x not in live]:
                    t.join()
                self.threads=live
            except selecterror:
                pass

    def HandleXMLCreatePG(self,xe,(peer,port)):
        # xe looks like
        # <create-process-group submitter='desai' pgid='1' totalprocs='32' output='discard'>
        #  <process-spec exec='/bin/foo' cwd='/etc' path='/bin:/usr/sbin' range='1-32'
        #              co-process='tv-server'>
        #     <arg idx='1' value='-v'/>
        #  </process-spec>
        #  <host-spec>
        #    node1
        #    node2
        #  </host-spec>
        # </create-process-group>
        if '-1' in argv:
            newpg=P4_ProcessGroup(xe)
        else:
            newpg=ProcessGroup(xe)
        newpg.pgid=self.GetPGID()
        self.pg[newpg.pgid]=newpg
        syslog(LOG_INFO, "proc-group %s created for %s"%(newpg.pgid,newpg.submitter))
        t=Thread(target=self.pg[newpg.pgid].Run,args=[],name=newpg.pgid)
        self.threads.append(t)
        t.start()
        # Deal with stdio
        creator=Document()
        response=creator.createElement("process-group")
        response.setAttribute('pgid',newpg.pgid)
        return response

    def HandleXMLPGSignal(self,xe,(peer,port)):
        sig=xe.getAttribute('signal')
        response=Document().createElement("process-groups")
        for pgrp in xe.getElementsByTagName('process-group'):
            pgid=pgrp.getAttribute('pgid')
            if pgid in self.pg.keys():
                self.pg[pgid].Signal(sig)
            else:
                syslog(LOG_INFO,"got stale pgid %s"%pgid)
            doc=Document().createElement("process-group")
            doc.setAttribute('pgid',pgid)
            response.appendChild(doc)
        return response

    def HandleXMLPGKill(self,xe,(peer,port)):
        response=Document().createElement("process-groups")
        for pgrp in xe.getElementsByTagName('process-group'):
            pgid=pgrp.getAttribute('pgid')
            if pgid in self.pg.keys():
                self.pg[pgid].Signal("SIGKILL")
            else:
                syslog(LOG_INFO,"got stale pgid %s"%pgid)
            doc=Document().createElement("process-group")
            doc.setAttribute('pgid',pgid)
            response.appendChild(doc)
        return response

    def HandleXMLPGInfo(self,xe,(peer,port)):
        creator=Document()
        response=creator.createElement("process-groups")
        # process xe
        for node in xe.childNodes:
            matches=filter(lambda x:x.Match(Node(node)), self.pg.values())
            map(lambda x:response.appendChild(x.toxmldom(Node(node))), matches)
        return response

if __name__ == '__main__':
    from getopt import getopt, GetoptError
    from os import fork
    from sys import argv,exit
    try:
        (opts,args)=getopt(argv[1:],'1dm:',['daemon='])
    except GetoptError,msg:
        print "%s\nUsage:\nmpdpm.py [-1] [-d] [-m <path to mpd> ] [--daemon <pidfile>]"%(msg)
        exit(1)
    daemon=filter(lambda x:(x[0] == '--daemon'),opts)
    mpath=filter(lambda x:(x[0] == '-m'),opts)
    ldebug=len(filter(lambda x:(x[0] == '-d'),opts))
    if mpath:
        mpdpath=mpath[0][1]
    if daemon:
        from sss.daemonize import daemonize
        daemonize(daemon[0][1])
    s=mpd(debug=ldebug)
    s.ServeForever()
