#!/usr/bin/env python

from cPickle import Pickler, Unpickler, dumps, loads
from getopt import getopt, GetoptError
from os import close,fork,pipe,read,write
from signal import signal, SIGINT
from sss.node import Node
from sss.server import SSSSelectServer, AddEvent, StateLoadError, StateSaveError
from sss.ssslib import comm_lib, InitError, ConnectError, CommunicationError
from string import find
from sys import argv,exc_info,exit
from syslog import syslog, LOG_INFO, LOG_ERR
from time import sleep
from traceback import extract_tb
from xml.dom.minidom import Document,parseString

class Location(Node):
    __attrfields__=['component','host','port','protocol','schema_version','tier']

    def __setdefaults__(self):
        self.tier='1'
        self.schema_version='1'

class service_directory(SSSSelectServer):
    __implementation__='sd'
    __component__='service-directory'
    __statefields__=['locations']

    def __parseinput__(self):
        self.dispatch={"add-location":self.XMLAddLocation,
                       "del-location":self.XMLDelLocation,
                       "get-location":self.XMLGetLocation}
        self.locations=[]
        try:
            self.LoadState()
        except StateLoadError:
            pass
        (port,host,proto,sversion,tier)=map(lambda x:comm_lib().GetConfig().get(x),['sd_port','sd_host','sd_protocol','sd_schema_version','sd_tier'])
        myloc='''<location component='service-directory' host='%s' port='%s' protocol='%s' schema_version='%s' tier='%s'/>'''%(host,port,proto,sversion,tier)
        self.AddLocation(Location(parseString(myloc).documentElement))
        self.port=int(port)
        self.protocol=proto
        
    def ChildSigHand(self,signum,frame):
        exit(0)

    def __setupChildren__(self):
        (self.pr,self.pw)=pipe()
        (self.cr,self.cw)=pipe()
        pid=fork()
        if pid == 0:
            signal(SIGINT, self.ChildSigHand)
            # in child
            close(self.cr)
            while 1:
                msglen=read(self.pr,8)
                msg=read(self.pr,int(msglen))
                try:
                    apply(AddEvent,loads(msg))
                except:
                    syslog(LOG_ERR, "Failed to send event")
        else:
            pinger_pid = fork()
            if pinger_pid == 0:
                self.RunPinger()
            else:
                close(self.pr)
                return [pid,pinger_pid]

    def RunPinger(self):
        try:
            c = comm_lib()
            while 1:
                sleep(300) # 5 minutes
                h = c.ClientInit("service-directory")
                c.SendMessage(h,"<get-location><location component='*' host='*' port='*' protocol='*'/></get-location>")
                locs = parseString(c.RecvMessage(h)).documentElement.getElementsByTagName("location")
                c.ClientClose(h)
                for loc in locs:
                    cmp = loc.getAttribute("component")
                    try:
                        h = c.ClientInit(cmp)
                        c.ClientClose(h)
                    except (ConnectError,CommunicationError):
                        h = c.ClientInit("service-directory")
                        c.SendMessage(h,"<del-location><location component='%s'/></del-location>"%cmp)
                        syslog(LOG_INFO,"component %s flushed from sd due to connect error"%cmp)
                        c.RecvMessage(h)
                        c.ClientClose(h)
                    except:
                        (t,v,tb)=exc_info()
                        syslog(LOG_ERR, "Failure in sd-pinger")
                        syslog(LOG_ERR, "Attempted to contact %s"%cmp)
                        for line in extract_tb(tb):
                            errstr = '  File "%s", line %i, in %s\n    %s\n'%line
                            syslog(LOG_ERR,errstr)
                            errstr = "%s: %s\n"%(t,v)
                            syslog(LOG_ERR,errstr)
        except KeyboardInterrupt:
            exit(1)
        except:
            (t,v,tb)=exc_info()
            syslog(LOG_ERR, "*** unexpected Failure in sd-pinger")
            for line in extract_tb(tb):
                errstr = '  File "%s", line %i, in %s\n    %s\n'%line
                syslog(LOG_ERR,errstr)
            errstr = "%s: %s\n"%(t,v)
            syslog(LOG_ERR,errstr)
            exit(1)
        
    def AddLocation(self,location):
        if not filter(lambda x:x.Match(location),self.locations):
            self.locations.append(location)
            try:
                self.SaveState()
            except StateSaveError:
                syslog(LOG_ERR, "Error saving state")

    # This is where the restriction stuff starts
    
    def XMLGetLocation(self,xml,(peer,port)):
        response=Document().createElement("locations")        
        for child in xml.childNodes:
            data=filter(lambda x:x.Match(Node(child)),self.locations)
            map(lambda x:response.appendChild(x.toxmldom(Node(child))),data)            
        # then return all requested data
        return response

    def XMLDelLocation(self,xml,(peer,port)):
        response=Document().createElement("locations")
        for child in xml.childNodes:
            data=filter(lambda x:x.Match(Node(child)),self.locations)
            map(lambda x:response.appendChild(x.toxmldom(Node(child))),data)
            self.locations=filter(lambda x:not x.Match(Node(child)),self.locations)
        try:
            self.SaveState()
        except StateSaveError:
            syslog(LOG_ERR, "Error saving state")
        return response

    def XMLAddLocation(self,xml,(peer,port)):
        response=Document().createElement("locations")
        for data in xml.childNodes:
            l=Location(data)
            self.AddLocation(l)
            response.appendChild(l.toxmldom(l))
            if data.getAttribute('component') != 'event-manager':
                msg=dumps(("service-directory","service-registered",data.getAttribute('component')))
                haswritten = 0
                while not haswritten:
                    try:
                        write(self.pw,"%08d"%(len(msg)))
                        haswritten = 1
                    except IOError, msg:
                        if find(str(msg),"Errno 4") > -1: # EINTR
                            pass        # restart the loop
                        else:
                            raise IOError,msg
                haswritten = 0
                while not haswritten:
                    try:
                        write(self.pw,msg)
                        haswritten = 1
                    except IOError, msg:
                        if find(str(msg),"Errno 4") > -1: # EINTR
                            pass        # restart the loop
                        else:
                            raise IOError,msg
        return response

    def __shutdown__(self):
        try:
            self.SaveState()
        except StateSaveError:
            syslog(LOG_ERR, "Error saving state")
    
if __name__ == "__main__":
    try:
        (opts,args)=getopt(argv[1:],'d',['daemon='])
    except GetoptError, msg:
        print "%s\nUsage:\nsd.py [-d] [--daemon <pidfile>]"%(msg)
        exit(1)
    daemon=filter(lambda x:(x[0] == '--daemon'),opts)
    ldebug=len(filter(lambda x:(x[0] == '-d'),opts))
    if daemon:
        from sss.daemonize import daemonize
        daemonize(daemon[0][1])
    try:
        s=service_directory(debug=ldebug)
    except InitError, msg:
        print "Library init Error:%s"%(msg)
        exit(1)
    s.ServeForever()
